// Maria temporal binary operator class -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "TemporalBinop.h"
#include "BoolType.h"
#include "Net.h"
#include "Property.h"
#include "Printer.h"

/** @file TemporalBinop.C
 * Temporal binary operators
 */

TemporalBinop::TemporalBinop (enum Op op,
			      class Expression& left,
			      class Expression& right) :
  Expression (),
  myOp (op), myLeft (&left), myRight (&right)
{
  assert (myLeft->getType ()->getKind () == Type::tBool);
  assert (myRight->getType ()->getKind () == Type::tBool);
  setType (Net::getBoolType ());
}

TemporalBinop::~TemporalBinop ()
{
  myLeft->destroy ();
  myRight->destroy ();
}

class Expression*
TemporalBinop::ground (const class Valuation& valuation,
		       class Transition* transition,
		       bool declare)
{
  class Expression* left = myLeft->ground (valuation, transition, declare);
  if (!left) return NULL;
  class Expression* right = myRight->ground (valuation, transition, declare);
  if (!right) { left->destroy (); return NULL; }

  assert (valuation.isOK ());

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else
    return (new class TemporalBinop (myOp, *left, *right))->cse ();
}

class Expression*
TemporalBinop::substitute (class Substitution& substitution)
{
  class Expression* left = myLeft->substitute (substitution);
  class Expression* right = myRight->substitute (substitution);

  if (!left || !right) {
    left->destroy ();
    right->destroy ();
    return NULL;
  }

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else
    return (new class TemporalBinop (myOp, *left, *right))->cse ();
}

bool
TemporalBinop::depends (const class VariableSet& vars,
			bool complement) const
{
  return
    myLeft->depends (vars, complement) ||
    myRight->depends (vars, complement);
}

bool
TemporalBinop::forVariables (bool (*operation)
			(const class Expression&,void*),
			void* data) const
{
  return
    myLeft->forVariables (operation, data) &&
    myRight->forVariables (operation, data);
}

class Ltl*
TemporalBinop::toFormula (class Property& property)
{
  enum Property::Binop op;
  switch (myOp) {
  case Until:
    op = Property::opUntil;
    break;
  case Release:
    op = Property::opRelease;
    break;
  }
  return property.addBinop (op, *myLeft, *myRight);
}

/** Convert an operator to a string
 * @param op	the operator to convert
 * @return	a string corresponding to the operator
 */
static const char*
getOpString (enum TemporalBinop::Op op)
{
  switch (op) {
  case TemporalBinop::Until:
    return "until";
  case TemporalBinop::Release:
    return "release";
  }

  return "???";
}

void
TemporalBinop::display (const class Printer& printer) const
{
  printer.delimiter ('(')++;
  myLeft->display (printer);
  --printer.delimiter (')');

  printer.printRaw (::getOpString (myOp));
  printer.delimiter ('(')++;
  myRight->display (printer);
  --printer.delimiter (')');
}
